/* -*- mode: c++; c-basic-offset: 3; indent-tabs-mode: nil; -*- */

#include <sys/types.h>
#include <sys/stat.h>      
#include <sys/time.h>   

// System Header Files   
#include <algorithm>   
#include <functional>
#include <iostream>   
#include <iterator>   
#include <sstream>
#include <stdexcept>   

#include <cerrno>

#include "general/Directory.hh"
#include "general/Thread.hh"

#include "genericAPI/Logging.hh"
#include "genericAPI/Stat.hh"
#include "general/System.hh"

// Local Header Files   
#include "cachefile.hh"   
#include "counter.hh"   
#include "dirhash.hh"   
#include "dirinfo.hh"
#include "diskcache.hh"
#include "diskcacheAPI.hh"
#include "filename.hh"
#include "HotDirectory.hh"
#include "mountpt.hh"   
#include "MountPointScanner.hh"
#include "timehash.hh"   

#if 0
#include "general/IOLock.hh"

#define VERBOSE_DEBUGGING 1
#define	AT_BASE( a ) \
{ \
   int oldstate; \
   pthread_setcancelstate( PTHREAD_CANCEL_DISABLE, &oldstate ); \
   MutexLock    lock( General::IOLock::GetKey( std::cerr ) ); \
   pthread_setcancelstate( oldstate, &oldstate ); \
\
   std::cerr << a << __FILE__ << " " << __LINE__ << std::endl << std::flush; \
}

#else
#define	AT_BASE( a )
#endif
#define	AT( a )	AT_BASE( a << std::endl << "\t" )
#define	HERE( ) AT_BASE( "" )

using namespace diskCache;
using namespace General;
using namespace std;
using GenericAPI::queueLogEntry;
using GenericAPI::LogEntryGroup_type;

// Local namespace
namespace
{
   
   // Useful typedefs   
   typedef list< string >::iterator LSIter;
   typedef list< string >::const_iterator LSConstIter;   

   //---------------------------------------------------------------------------
   //
   /// \brief Get errno string message.
   ///   
   /// This method interpret error code into the string message.
   ///
   /// \return
   ///     Formatted error message.
   //---------------------------------------------------------------------------

   inline const string
   errnoMessage( )
   {
      ostringstream msg;
      msg << " errno=" << errno << " (";
   
      string errno_msg( strerror( errno ) );
      msg << errno_msg << ')';
   
   
      return msg.str();
   }

   inline void
   do_stat( const std::string& Path, struct stat& StatStruct )
   {
      static const CHAR* const method_name( "do_stat@dirinfo.cc: " );

      int err = GenericAPI::LStat( Path, StatStruct );

      AT( "err: " << err );
      if ( StdErrLog.IsOpen( ) )
      {
        HERE( );
	std::ostringstream	msg;
      
	msg << "doStat: " << Path
	    << " err: " << err
	    << " errno: " << errno;
	StdErrLog( General::ErrorLog::DEBUG,
		   __FILE__, __LINE__,
		   msg.str( ) );
      }
      if( err != 0 )
      {
         //--------------------------------------------------------------
         // An error has occurred while trying to get information about
         // this directory.
         // 1) Create% a readable error message
         //--------------------------------------------------------------
         HERE( );
         string exc( method_name );
         exc += "Could not execute lstat() for \'";
         exc += Path;
         exc += '\'';
         exc += errnoMessage( );
         exc += " ";

         //--------------------------------------------------------------
         /// \todo
         ///     Remove all subdirectories recursively
         //--------------------------------------------------------------
         throw runtime_error( exc );
      }
   }
}

  
//------------------------------------------------------------------------------
//   
//: Constructor   
//
//!param: const std::string& path - A full path of directory.
//!param: const std::string& mount_pt - A full path for the mount point 
//+       directory resides under.
//   
diskCache::DirectoryInfo::
DirectoryInfo( const std::string& path, 
               const std::string& mount_pt )   
   : m_scan_baton( MutexLock::Initialize( ) ),
     mPath( path ),
     mMountPtPath( mount_pt ),
     m_file_count( 0 ),
     m_last_modified( MODIFIED_NEW_DIRECTORY ),
     p_index_baton( General::ReadWriteLock::Initialize( ) )
{
   ReadWriteLock::Initialize( mLock );
   ReadWriteLock::Initialize( m_const_lock );
}
   

//------------------------------------------------------------------------------
//   
//: Destructor   
//
diskCache::DirectoryInfo::
~DirectoryInfo()
{
#if WORKING
   {
      ReadWriteLock lock( mLock, ReadWriteLock::WRITE,
                          __FILE__, __LINE__ );   
      ReadWriteLock cbaton( m_const_lock, ReadWriteLock::WRITE,
                            __FILE__, __LINE__ );   
      HotDirectory::Unregister( this );

      get_index( ).Var( ).clear( );
   }
#endif /* WORKING */

   
   ReadWriteLock::Destroy( m_const_lock );
   ReadWriteLock::Destroy( mLock );
}
   
   
//------------------------------------------------------------------------------
/// This method is used for searching for a subset of the information.
///
/// \todo Evaluate diskCache::DirectoryInfo::Find for inlining
//------------------------------------------------------------------------------
void diskCache::DirectoryInfo::
Find( const Query& SearchParams,
      QueryResults& Results ) const
{
   const index_variable_ro_type i( get_index( ) );

   i.Var( ).Find( SearchParams, Results );
}

//------------------------------------------------------------------------------
//   
//: Scan directory.
//
// This method scans all subdirectories and sublinks of the directory
// (recursively).
//   
//!param: const std::list< std::string >& exclude_dirs - A list of known 
//+       directory names that should not be scanned.   
//!param: std::string& errors - An accumulative error message.
//!param: const INT_4S new_mtime - Newly acquired mtime for the directory.
//   
//!return: const std::list< std::string > - List of identified/scanned
//+        subdirectories with frame data.
//
std::list< std::string > diskCache::DirectoryInfo::
scan( const std::list< std::string >& exclude_dirs, 
      LogMessage& errors,
      bool& Modified )
{
   MutexLock    scan_lock( m_scan_baton );

   static const CHAR* const method_name( "DirectoryInfo::scan( ): " );
   std::list< std::string > retval;
   Modified = false;

   //--------------------------------------------------------------------
   // Be cooperative with cancelation process
   //--------------------------------------------------------------------
   Thread::Self::CancellationCheck( "diskCache::DirectoryInfo::scan",
                                    __FILE__, __LINE__ );


   //--------------------------------------------------------------------
   /// \todo
   ///     Validate the path of this entry.
   //--------------------------------------------------------------------

   try
   {
      struct stat stat_struct;

      const std::string path( GetPath( ).Var( ) );

      do_stat( path, stat_struct );
      AT( "Dir modification time: " << stat_struct.st_mtime << " Previous modification time: " << TimeModified( ) );
      //-----------------------------------------------------------------
      // Successful with getting the file's status.
      // Now verify other conditions
      //-----------------------------------------------------------------
      if ( TimeModified( ) == stat_struct.st_mtime )
      {
         //--------------------------------------------------------------
         // If the time stamp has not changed, then there is nothing
         //   to be done.
         //--------------------------------------------------------------
         ScannedFiles( m_file_count );
         return retval;
      }
      if ( S_ISLNK( stat_struct.st_mode ) )
      {
         //--------------------------------------------------------------
         // Any entry that is a symbolic link is an error
         //--------------------------------------------------------------
         HERE( );
         string exc( method_name );
         exc += "API does not support symbolic links. Please remove \'";
         exc += path;
         exc += '\'';
         
         throw runtime_error( exc );         
      }
      if ( S_ISDIR( stat_struct.st_mode ) == 0 )
      {
         HERE( );
         if( path == GetMountPointPath( ).Var( ) )
         {
            string exc( method_name );
            exc += "Specified path \'";
            exc += path;
            exc += "\' is not a directory.";
         
            throw runtime_error( exc );      
         }
         
         //--------------------------------------------------------------
         // Just a file (other than frame file) in directory being examined ==> 
         // ignore it (frame files are captured in readDirectory() by 
         // extension match)
         //--------------------------------------------------------------
         AT( method_name << ": Exiting" );
   
         return retval;
      }
      const bool      is_new( m_last_modified == MODIFIED_NEW_DIRECTORY );
      INT_4U    added_file_count( 0 );
      INT_4U    removed_file_count( 0 );
      General::SharedPtr< DirectoryInfo >       new_dir_info;
      {
         //--------------------------------------------------------------
         // Lock because things are going to be modified below here
         //--------------------------------------------------------------
         General::ReadWriteLock    l( mLock,  General::ReadWriteLock::WRITE,
                                      __FILE__, __LINE__ );
         //--------------------------------------------------------------
         // Read the contents of the directory.
         //--------------------------------------------------------------
         new_dir_info.reset( new DirectoryInfo( path, GetMountPointPath( ).Var( ) ) );
         new_dir_info->read_directory( exclude_dirs,
                                       errors,
                                       *this,
                                       new_dir_info,
                                       added_file_count,
                                       removed_file_count );
         new_dir_info->TimeModified( stat_struct.st_mtime );
         ScannedFiles( new_dir_info->m_file_count );
      }
      //-----------------------------------------------------------------
      // Commit changes
      //-----------------------------------------------------------------
      DirectoryHash::Instance( ).AddDirectory( new_dir_info );
      Modified = true;
      //-----------------------------------------------------------------
      // Generate message
      //-----------------------------------------------------------------
      if ( removed_file_count )
      {
         errors( LogMessage::LM_NOTE )
            <<" {"
            << new_dir_info->GetPath( ).Var( )
            << ( ( new_dir_info->m_file_count )
                 ? " UPDATED"
                 : " REMOVED" )
            << " DIRECTORY WITH "
            << removed_file_count
            << " FILE"
            << ( ( removed_file_count == 1 )
                 ? ""
                 : "S" )
            << " REMOVED} "
            ;
      }      
      if ( added_file_count )
      {
         errors( LogMessage::LM_NOTE )
            <<" {"
            << new_dir_info->GetPath( ).Var( )
            << ( ( is_new )
                 ? " NEW"
                 : " UPDATED" )
            << " DIRECTORY WITH "
            << added_file_count
            << " FILE"
            << ( ( added_file_count == 1 )
                 ? ""
                 : "S" )
            << " ADDED} "
            ;
      }      
   }
   catch( const General::Thread::cancellation& exc )
   {
      throw;
   }
   catch( const exception& exc )
   {
      //-----------------------------------------------------------------
      // Need to setup to scan again.
      //-----------------------------------------------------------------
      TimeModifiedReset( );

      errors( LogMessage::LM_NOTE ) << " {"
                                    << exc.what()
                                    << "} "
         ;
   }

   return list< string >();
}

   
//------------------------------------------------------------------------------
//
//: Get Hash Numbers for "this" directory
//   
//!param: const std::string& ifo_type - A specific ifo and type pair. 
//+       Default is empty string: get all available data information.
//   
//!return: const std::string - A list of all identified subdirectories
//+        with frame data with a Tcl-layer requested metadata.
// 
const int diskCache::DirectoryInfo::
getHashNumbers( const std::string& ifo_type, bool TryLock ) const
{
   ReadWriteLock lock( mLock, ReadWriteLock::READ, TryLock,
                       __FILE__, __LINE__ );

   // number of files for "this" directory
   int number_of_files(0);

#if old
   if( mFrameMap.empty() == true )
   {
      return 0;
   }

   // Data of specific ifo and type is requested
   if( ifo_type.empty() == false )
   {
       MSConstIter iter = mFrameMap.find( ifo_type );
       if( iter == mFrameMap.end() )
       {
          return 0;
       }
       number_of_files = getHashNumbersHelper( iter->second );
   } 
   // All ifo-type data requested
   else  
   {
       for( MSConstIter iter = mFrameMap.begin(), end_iter = mFrameMap.end();
            iter != end_iter; ++iter )
       {
            number_of_files += getHashNumbersHelper( iter->second );
       }
   }
#endif /* old */
    
   return number_of_files;
}   

//------------------------------------------------------------------------------
//
//: Get Frame Query Intervals for "this" directory.
//   
// This method gets ASCII representation of frame data hash 
// for "this" directory in Tcl-layer requested format.   
//
//!param: const std::string& ifo_type - A specific ifo and type pair. 
//+       Default is empty string: get all available data information.
//!param: const unsigned int start_time - query start time.
//!param: const unsigned int stop_time - query stop time.
//   
//!return: const std::string - A KnownIntervals object with intervals for this
//+                            directory within the bounds of start, stop
// 
const TimeHash::KnownIntervals diskCache::DirectoryInfo::
getFrameQueryIntervals( const std::string& ifo_type,
                        const INT_4U query_start, 
                        const INT_4U query_stop ) const
{
   ReadWriteLock lock( mLock, ReadWriteLock::READ, RWLOCK_TIMEOUT,
                       __FILE__, __LINE__ );

   TimeHash::KnownIntervals empty_interval;
   
#if old
   if( mFrameMap.empty() == true )
   {
      return empty_interval;
   }

   // Data of specific ifo and type is requested
   if( ifo_type.empty() == false )
   {
       MSConstIter iter = mFrameMap.find( ifo_type );
     
       // return empty vector if data is not found
       if( iter == mFrameMap.end() )
       {
          return empty_interval;
       }

       // iter->second is a TimeHash object from the mFrameMap
       // call its getFrameQueryIntervals method with 
       // this directory's path, ifo_type, start, and stop

       return ( (iter->second)->getFrameQueryIntervals(
                                   mPath, ifo_type, query_start, query_stop ) );
   } 
#endif /* old */

   return empty_interval;
}   


   
//------------------------------------------------------------------------------
//
//: Get list of identified subdirectories from previous scan.
//   
//!return: const std::list< std::string > - A list of all subdirectories.
// 
const std::list< std::string > diskCache::DirectoryInfo::getSubDirectoryList() 
   const
{
   ReadWriteLock lock( mLock, ReadWriteLock::READ, RWLOCK_TIMEOUT,
                       __FILE__, __LINE__ );            
   return mSubDirs;
}

   
//------------------------------------------------------------------------------
//
//: Get full path of the directory.
//   
//!return: const std::string - A directory's full path.
// 
const std::string diskCache::DirectoryInfo::getPath( bool TryLock ) const
{
   ReadWriteLock lock( mLock, ReadWriteLock::READ, TryLock,
                       __FILE__, __LINE__ );            
   return mPath;
}

   
//------------------------------------------------------------------------------
//
//: Is directory a MOUNT_PT entry.
//   
//!return: const bool - True if directory is a MOUNT_PT entry, false otherwise.
// 
const bool diskCache::DirectoryInfo::isMountPtEntry() const
{
   ReadWriteLock lock( mLock, ReadWriteLock::READ, RWLOCK_TIMEOUT,
                       __FILE__, __LINE__ );            
   return ( mPath == mMountPtPath );
}

   
//------------------------------------------------------------------------------
//
//: Get a MOUNT_PT entry for the directory
//   
//!return: const std::string - MOUNT_PT entry directory belongs to.
// 
const std::string diskCache::DirectoryInfo::getMountPtPath() const
{
   try
   {
   ReadWriteLock lock( mLock, ReadWriteLock::READ, RWLOCK_TIMEOUT,
                       __FILE__, __LINE__ );            

   return mount_point( );
   }
   catch( ... )
   {
      throw;
   }
}
   
   
//------------------------------------------------------------------------------
//
//: Get time hash for a specific "IFO-Type-" in TimeHash::KnownIntervals format.
//   
//!return: const TimeHash::KnownIntervals - A time hash information.
//    
const TimeHash::KnownIntervals diskCache::DirectoryInfo::getKnownIntervals( 
   const std::string& ifo_type ) const
{
   TimeHash::KnownIntervals known_data;
   
   // Get a copy of internal time hash
   {   
      ReadWriteLock lock( mLock, ReadWriteLock::READ, RWLOCK_TIMEOUT,
                          __FILE__, __LINE__ );      
#if old
      MSConstIter iter( mFrameMap.find( ifo_type ) );
   
      if( iter != mFrameMap.end() )
      {
         // "IFO-Type-" is registered with a map
         known_data.setIntervals( iter->second );
      }
#endif /* old */
   }

   return known_data;
}

   
//------------------------------------------------------------------------------
//
//: Write object to the stream.
//   
//!param: CacheFile& stream - A reference to the stream to write object to.
//   
//!return: Nothing.
// 
void diskCache::DirectoryInfo::
write( CacheFile& stream ) const
{
   ReadWriteLock lock( mLock, ReadWriteLock::READ, RWLOCK_TIMEOUT,
                       __FILE__, __LINE__ );                
   
   
   const index_variable_ro_type i( get_index( ) );
   
   stream << mPath
          << mMountPtPath
          << mSubDirs 
          << mExcludedSubDirs
          << m_last_modified
          << i.Var( )
      ;
   
   
   return;
}
   
   
//------------------------------------------------------------------------------
//
//: Read object from the stream.
//   
//!param: CacheFile& stream - A reference to the stream to read object from.
//   
//!return: DirectoryInfo* const - A pointer to the newly read object.
// 
// static
General::SharedPtr< DirectoryInfo > diskCache::DirectoryInfo::
read( CacheFile& stream )
{
   static const char func_name[] = "diskCache::DirectoryInfo::read";

   string dir, mount_pt;
   
   // Read directory full path and mount point it resides under
   stream >> dir >> mount_pt;

   if ( DISKCACHE__CACHE_FILE_VERBOSE_DEBUGGING )
   {
      std::ostringstream   msg;

      msg << "Reading: dir: " << dir
          << " mount_pt: " << mount_pt
         ;
   
      queueLogEntry( msg.str( ),
                     LogEntryGroup_type::MT_DEBUG,
                     20,
                     func_name,
                     "CXX_INTERFACE" );
   }
                  
   General::SharedPtr< DirectoryInfo > new_dir( new DirectoryInfo( dir, mount_pt ) );
   
   // Read the rest of the directory metadata
   new_dir->readInfo( stream, new_dir );

   //----------------------------------------------------------------------------
   // Registration happens in diskCache::DirectoryHash::readDirectories
   //----------------------------------------------------------------------------

   return new_dir;
}
   
   
//------------------------------------------------------------------------------
/// This method reads a CacheFile stream so as to reconstruct the internals
/// of DirectoryInfo object.
///
/// \todo The DirInfo parameter should be removed as this object should
///     only be created and referenced using a shared_ptr class.
///     To be able to get rid of the DirInfo parameter, the TimeHash
///     read method needs to be modified to take a pointer to a
///     DirectoryInfo object.
//------------------------------------------------------------------------------
void diskCache::DirectoryInfo::
readInfo( CacheFile& stream,
          General::SharedPtr< DirectoryInfo > DirInfo )
{
   static const char caller[] = "diskCache::DirectoryInfo::readInfo";

   ReadWriteLock lock( mLock, ReadWriteLock::WRITE,
                       __FILE__, __LINE__ );            
   
   stream
      >> mSubDirs               // Read list of subdirectories
      >> mExcludedSubDirs       // Read list of excluded subdirectories
      >> m_last_modified        // Read time directory was last modified
      ;
   get_index( ).Var( ).read( stream, DirInfo );    // Read compressed hash information

   m_file_count = get_index( ).Var( ).FileCount( );
   ScannedFiles( m_file_count );

   if ( DISKCACHE__CACHE_FILE_VERBOSE_DEBUGGING )
   {
      std::ostringstream        msg;

      msg << "Read: {"
         ;
      for ( std::list< std::string >::const_iterator
               cur = mSubDirs.begin( ),
               last = mSubDirs.end( );
            cur != last;
            ++cur )
      {
         msg << " " << *cur;
      }
      msg << " } { "
         ;
      for ( std::list< std::string >::const_iterator
               cur = mExcludedSubDirs.begin( ),
               last = mExcludedSubDirs.end( );
            cur != last;
            ++cur )
      {
         msg << " " << *cur;
      }
      msg << " } " << m_last_modified
          << " <m_index> }"
          << " m_file_count: " << m_file_count
         ;

      queueLogEntry( msg.str( ),
                     LogEntryGroup_type::MT_DEBUG,
                     30,
                     caller,
                     "CXX_INTERFACE" );
   }

   return;
}
   
   
//------------------------------------------------------------------------------
//
//: Set subdirectory as excluded directory.
//   
// This method registers subdirectory in excluded directories list. This list
// is determined by MountPointHash::mExcludedDirList.   
//   
//!return: Nothing.
// 
void diskCache::DirectoryInfo::setExcludedSubDir( const std::string& sdir )
{
   ReadWriteLock lock( mLock, ReadWriteLock::WRITE,
                       __FILE__, __LINE__ );            
   
   mSubDirs.remove( sdir );

   
   if( find( mExcludedSubDirs.begin(), mExcludedSubDirs.end(), sdir ) == 
       mExcludedSubDirs.end() )
   {
      mExcludedSubDirs.push_back( sdir );
   }
   
   
   return;
}
   
//-----------------------------------------------------------------------
// This method unregisters all subdirectories "mPath" has introduced from
// global hash (recursively). 
//
// It gets called by DirectoryHash::removeDirectory() method only: 
// it's guaranteed that "this" will be deleted once
// removeSubDirectories() method returns.
//-----------------------------------------------------------------------
void diskCache::DirectoryInfo::
removeSubDirectories( dir_container_type& RemovedDirectories,
                      LogMessage& errors,
                      const bool unregister )
{
   // It's safe to lock "this" object while descending through all
   // subdirectories: "this" directory is already removed from global hash.
   ReadWriteLock lock( mLock, ReadWriteLock::WRITE,
                       __FILE__, __LINE__ );      

   list< string > removed_subdirs;
   
   // Number of files removed from diskcache
   int num_files_removed = 0;

#if WORKING
   // Include directory itself if it contains frame data
   // Kind of redundant check: mEntry != 0; just to be safe    
   if( mEntry && mEntry->hasFrameData() )
   {
      removed_subdirs.push_back( mPath );
      num_files_removed = mEntry->getRegularFileCount( );
   }
#endif /* WORKING */

   // Remove references to "this" object from MOUNT_PT entry in frame hash
   if( unregister )
   {
      get_index( ).Var( ).Unregister( errors );
      // Destructor for "this" object will free TimeHash* objects in the map
   }
   
   
   // Cleanup subdirs
   for( LSIter
           iter = mSubDirs.begin(),
           end_iter = mSubDirs.end();
        iter != end_iter;
        ++iter )
   {
      DirectoryHash::Instance().removeDirectory( *iter,
                                                 RemovedDirectories,
                                                 errors,
                                                 unregister );
   }

   // FIX FIX 
   // Append 'REMOVED' message to errors

   // Send REMOVE messages back to tcl layer via errors string
   errors( LogMessage::LM_NOTE )
      <<" {"
      << mPath
      << " REMOVED DIRECTORY WITH "
      << num_files_removed
      << ( ( num_files_removed == 1 )
           ? " FILE REMOVED} "
           : " FILES REMOVED} " )
      ;
   RemovedDirectories.splice( RemovedDirectories.end( ), removed_subdirs );
}
   
//------------------------------------------------------------------------------
/// Read the contents of the directory and put the data into meaningful
/// lists.
//------------------------------------------------------------------------------
void diskCache::DirectoryInfo::
read_directory( const std::list< std::string >& ExcludedDirectories,
                LogMessage& errors,
                const DirectoryInfo& PreviousInfo,
                General::SharedPtr< DirectoryInfo > CurrentInfo,
                INT_4U& AddedFileCount,
                INT_4U& RemovedFileCount )
{
   static const char func_name[] = "diskCache::DirectoryInfo::read_directory";

   static const CHAR* const method_name( "diskCache::DirectoryInfo::read_directory: " );

   //--------------------------------------------------------------------
   // Lock because things are going to be modified below here
   //--------------------------------------------------------------------
   General::ReadWriteLock    l( mLock,  General::ReadWriteLock::WRITE,
                                __FILE__, __LINE__ );
   
   //--------------------------------------------------------------------
   // Get directory content
   //--------------------------------------------------------------------
   HERE( );
   std::string file_path;
   General::Directory        current_dir( GetPath( ).Var( ) );

   file_path = current_dir.Name( );
   file_path += '/';

   FilenameHandler::IsFileData  fd( PreviousInfo, CurrentInfo );

   //--------------------------------------------------------------------
   // Collect all frame files in the directory, then figure out 
   //--------------------------------------------------------------------
   try
   {
      current_dir.Open( );
      while( current_dir.Next( ) )
      {
         AT( "current_dir.EntryName( ): "
             << current_dir.EntryName( ) );

         FileName current_file( current_dir.EntryName( ), file_path );
   
      
         //-----------------------------------------------------------
         // Skip entries that start with '.' ('.', '..', '.XXX'), 
         // and temporary frame files
         //-----------------------------------------------------------
         if( current_file.shouldIgnore() )
         {
            continue;
         }
       
         //-----------------------------------------------------------
         // By Tcl request: just assume that any file name with
         // frame extension is a frame file ===> don't run doStat() on it
         // 
         // Note: isFrame() will call parse() on file 
         //-----------------------------------------------------------
            
         if ( FilenameHandler::IsFileCheck( current_dir.EntryName( ), fd ) )
         {
            //--------------------------------------------------------
            // Someone recorded the filename so it is assumed to
            // be a regular file and not a directory.
            //-----------------------<---------------------------------
            AT( "IsFileCheck: true EntryName: " << current_dir.EntryName( ) );
            continue;
         }

         //-----------------------------------------------------------
         // Format full path for the file
         //-----------------------------------------------------------
         string temp_file( file_path );
         temp_file += current_dir.EntryName( );
   
         //-----------------------------------------------------------
         // If entry should be excluded from the scan
         //-----------------------------------------------------------
         if( current_file.shouldExclude( ExcludedDirectories ) )
         {
            mExcludedSubDirs.push_back( current_dir.EntryName( ) );
            continue;
         }
         
         HERE( );
         try
         {
            struct stat stat_struct;

            do_stat( temp_file, stat_struct );

            if ( S_ISLNK( stat_struct.st_mode ) )
            {
               //--------------------------------------------------------
               // Any entry that is a symbolic link is an error
               //--------------------------------------------------------
               HERE( );
               string exc( method_name );
               exc += "API does not support symbolic links. Please remove \'";
               exc += temp_file;
               exc += '\'';
         
               throw runtime_error( exc );         
            }
            if ( S_ISDIR( stat_struct.st_mode ) == 0 )
            {
               HERE( );
               if( temp_file == GetMountPointPath( ).Var( ) )
               {
                  string exc( method_name );
                  exc += "Specified path \'";
                  exc += temp_file;
                  exc += "\' is not a directory.";
         
                  throw runtime_error( exc );      
               }
               
               //--------------------------------------------------------
               // Just a file (other than frame file) in directory being
               // examined ==> ignore it (frame files are captured in
               // readDirectory() by extension match)
               //--------------------------------------------------------
               continue;
            }
            //-----------------------------------------------------------
            // Add the directory to list of known directories.
            /// \todo
            ///     Should optimise by preserving the directory name
            ///     relative to the parent directory.
            //-----------------------------------------------------------
            // mSubDirs.push_back( current_dir.EntryName( ) );
            mSubDirs.push_back( temp_file );
         }
         catch( const General::Thread::cancellation& exc )
         {
            throw;
         }
         catch( const exception& exc )
         {
            //-----------------------------------------------------------
            // Need to setup to scan again.
            //-----------------------------------------------------------
            TimeModifiedReset( );

            errors( LogMessage::LM_NOTE ) << " {"
                                          << exc.what()
                                          << "} "
               ;
         }
      }
      current_dir.Close( );
      //-----------------------------------------------------------------
      // Sort the subdirectories
      //-----------------------------------------------------------------
      mSubDirs.sort( );
      //-----------------------------------------------------------------
      // Remove any directories that have been removed.
      //-----------------------------------------------------------------
      {
         INT_4U                 n( 0 );
         std::list< string >    removed;

         std::list< string >::const_iterator
            cur, last;

         if ( ( PreviousInfo.mSubDirs.size( ) > 0 )
              && ( mSubDirs.size( ) > 0 ) )
         {
            insert_iterator< std::list< string > > remove_iter( removed, 
                                                                removed.begin( ) );

            set_difference( PreviousInfo.mSubDirs.begin( ),
                            PreviousInfo.mSubDirs.end( ),
                            mSubDirs.begin( ),
                            mSubDirs.end( ),
                            remove_iter );
            cur = removed.begin( );
            last = removed.end( );
            n = removed.size( );
         }
         else if ( mSubDirs.size( ) == 0 )
         {
            cur = PreviousInfo.mSubDirs.begin( );
            last = PreviousInfo.mSubDirs.end( );
            n = PreviousInfo.mSubDirs.size( );
         }
         else
         {
            cur = removed.begin( );
            last = removed.end( );
            n = removed.size( );
         }
         //--------------------------------------------------------------
         // \todo 
         //     Generate a list of directories that have been removed.
         //--------------------------------------------------------------
#if WORKING
         std::cerr << "DEBUG:  removed: " << n
                   << " directories"
                   << std::endl
            ;
#endif /* WORKING */
         //--------------------------------------------------------------
         // Remove the directories
         //--------------------------------------------------------------
         for ( ;
               cur != last;
               ++cur )
         {
            DirectoryHash::Instance( ).RemoveDirectory( *cur );
         }
      }
      //-----------------------------------------------------------------
      // Calculate the number of directories that have been added.
      //-----------------------------------------------------------------
      {
         INT_4U n = mSubDirs.size( );
         if ( ( PreviousInfo.mSubDirs.size( ) > 0 )
              && ( mSubDirs.size( ) > 0 ) )
         {
            std::list< string >    added;

#if WORKING
            std::cerr << "DEBUG: previous size: " << PreviousInfo.mSubDirs.size( )
                      << " current: " << mSubDirs.size( )
                      << std::endl
               ;
#endif /* WORKING */
            insert_iterator< std::list< string > > added_iter( added, 
                                                               added.begin( ) );

            set_difference( mSubDirs.begin( ),
                            mSubDirs.end( ),
                            PreviousInfo.mSubDirs.begin( ),
                            PreviousInfo.mSubDirs.end( ),
                            added_iter );
            n = added.size( );
         }
         //--------------------------------------------------------------
         // \todo 
         //     Generate a list of directories that have been removed.
         //--------------------------------------------------------------
#if WORKING
         std::cerr << "DEBUG:  added: " << n
                   << " directories"
                   << std::endl
            ;
#endif /* WORKING */
      }
      //-----------------------------------------------------------------
      // Finish processing anything that the cache algorithms need to do.
      //-----------------------------------------------------------------
      fd.OnDirectoryClose( );
      AddedFileCount = fd.FilesAdded( );
      RemovedFileCount = fd.FilesRemoved( );
      for ( diskCache::FilenameHandler::IsFileData::private_data_container_type::const_iterator
               ifd_cur = fd.s_private_data.begin( ),
               ifd_last = fd.s_private_data.end( );
            ifd_cur != ifd_last;
            ++ifd_cur )
      {
         for ( diskCache::FilenameHandler::IsFileData::conflict_container_type::const_iterator
                  conflict_cur = ifd_cur->second->s_conflict_messages.begin( ),
                  conflict_last = ifd_cur->second->s_conflict_messages.end( );
               conflict_cur != conflict_last;
               ++conflict_cur )
         {
            std::string    msg( "CONFLICT: " );
            msg += *conflict_cur;
            
            queueLogEntry( msg,
                           LogEntryGroup_type::MT_EMAIL,
                           0,
                           func_name,
                           "SCAN_MOUNTPT" );
         }
      }
      //-----------------------------------------------------------------
      // Calculate the number of files that are being managed by this
      // object.
      //-----------------------------------------------------------------
      m_file_count = get_index( ).Var( ).FileCount( );
   }
   catch( ... )
   {
      current_dir.Close( );
      throw;
   }

}
//
// END 
//
